"""Methods that wrap the FFIEC Webservice API

The methods contained in this module are utilized to call and collect data from the FFIEC Webservice API.

"""

import re
import requests
import pandas as pd
from typing import Union
from datetime import datetime
from zoneinfo import ZoneInfo
from zeep import Client, Settings
from zeep.wsse.username import UsernameToken
from zeep.transports import Transport
from ffiec_data_connect import datahelpers, credentials, constants, xbrl_processor, ffiec_connection

# global date regex
quarterStringRegex = r"^[1-4](q|Q)([0-9]{4})$"
yyyymmddRegex = r"^[0-9]{4}[0-9]{2}[0-9]{2}$"
yyyymmddDashRegex = r"^[0-9]{4}-[0-9]{2}-[0-9]{2}$"
mmddyyyyRegex = r"^[0-9]{1,2}/[0-9]{1,2}/[0-9]{4}$"

validRegexList = [quarterStringRegex, yyyymmddRegex, yyyymmddDashRegex, mmddyyyyRegex]


def _create_ffiec_date_from_datetime(indate: datetime) -> str:
    """Converts a datetime object to a FFIEC-formatted date

    Args:
        indate (datetime): the date to convert

    Returns:
        str: the date in FFIEC format
    """
    month_str = str(indate.month)
    day_str = str(indate.day)
    year_str = str(indate.year)
    
    mmddyyyy = month_str + "/" + day_str + "/" + year_str
    
    return mmddyyyy

def _convert_any_date_to_ffiec_format(indate: str or datetime) -> str:
    """Converts a string-based date or python datetime object to a FFIEC-formatted date
 
    Args:
        date (str or datetime): the date to convert. This can be a string in the format of "YYYY-MM-DD", "YYYYMMDD", "MM/DD/YYYY", or a python datetime object

    Returns:
        str: the date in FFIEC format
    """
    
    if isinstance(indate, datetime):
        return _create_ffiec_date_from_datetime(indate)
    elif isinstance(indate, str):
        # does the date have two slashes?
        if indate.count("-") == 2:
            return _create_ffiec_date_from_datetime(datetime.strptime(indate, "%Y-%m-%d"))
        elif indate.count("/") == 2:
            return _create_ffiec_date_from_datetime(datetime.strptime(indate, "%m/%d/%Y"))
        elif len(indate) == 8:
            return _create_ffiec_date_from_datetime(datetime.strptime(indate, "%Y%m%d"))  
    else:
        # raise an error if we don't have a valid date
        raise(ValueError("Invalid date format. Must be a string in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', or a python datetime object"))
    
def _convert_quarter_to_date(reporting_period: str) -> datetime:
    
    """Converts date in the format of #QYYYY to a datetime object

    Returns:
        _type_: _description_
    """
    
    # convert the reporting period to a datetime object
    if re.search(quarterStringRegex, reporting_period):
        # the reporting period is a quarter string
        # get the quarter number
        quarter_number = int(reporting_period[0])
        # get the year
        year = int(reporting_period[-4:])

        if quarter_number == 1:
            # first quarter
            return datetime(year, 3, 31)
        elif quarter_number == 2:
            return datetime(year, 6, 30)
        elif quarter_number == 3:
            return datetime(year, 9, 30)
        elif quarter_number == 4:
            return datetime(year, 12, 31)
        else:
            raise(ValueError("Invalid quarter number")) # raise an error if we don't have a valid quarter number
    pass
    

def _is_valid_date_or_quarter(reporting_period: str or datetime) -> bool:
    
    """ Validates the reporting period input argument, which should indicate either the name of a calendar quarter, or a string that represents the last day of a quarter (e.g. "2019-03-31"), or a datetime object.
    
    If reporting period is a datetime, validate that the date is at quarter end.
    
    If reporting period is a string, validate that the string is in the format of "Q#-YYYY", "Q#-YY", "YYYY-MM-DD", "YYYYMMDD", or m/d/YYYY, or m/d/YY.
    
    Args:
        reporting_period (str or datetime): the reporting period to validate

    Returns:
        bool: True if valid reporting period, False if not valid reporting period
    
    """

    if isinstance(reporting_period, datetime):
        # what is the month of the quarter?
        month = reporting_period.month # 1 = Jan, 12= Dec
        day = reporting_period.day # 1 = 1st
        
        if month in [3,12]:
            if day == 31:
                return True # the quarter ends on the 31st in March and December
            else:
                return False
        elif month in [6,9]:
            if day == 30:
                return True # the quarter ends on the 30th in June, September
            else:
                return False
    elif isinstance(reporting_period, str):
        # does our date match any of the valid regexes?
        return any(re.search(regex, reporting_period) for regex in validRegexList)
    else:
        return False # we don't know what to do with this type of input, so return false
    
    
def _return_ffiec_reporting_date(indate: datetime or str) -> str:
    if isinstance(indate, datetime):
        return _create_ffiec_date_from_datetime(indate)
    elif isinstance(indate, str):
        if indate[1] == "Q":
            return _create_ffiec_date_from_datetime(_convert_quarter_to_date(indate))
        else:
            ffiec_date =  _convert_any_date_to_ffiec_format(indate)
            ffiec_date_month = ffiec_date.split("/")[0]
            ffiec_date_date = ffiec_date.split("/")[1]
            
            if (ffiec_date_month == "3" or ffiec_date_month == "03") and ffiec_date_date == "31":
                return ffiec_date
            elif (ffiec_date_month == "6" or ffiec_date_month == "06") and ffiec_date_date == "30":
                return ffiec_date
            elif (ffiec_date_month == "9" or ffiec_date_month == "09") and ffiec_date_date == "30":
                return ffiec_date
            elif ffiec_date_month == "12" and ffiec_date_date == "31":
                return ffiec_date
            else:
                raise(ValueError("Invalid date format. Must be a string in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', or a python datetime object"))
    
def _output_type_validator(output_type: str) -> bool:
    """Internal function to validate the output_type
    
    Args:
        output_type (str): the output_type to validate
    
    Returns:
        bool: True if valid, False if not
    """
    if output_type not in ['list', 'pandas']:
        raise(ValueError("Invalid output_type. Must be 'list' or 'pandas'"))
    else:
        return True
    
def _date_format_validator(date_format: str) -> bool:
    """Internal function to validate the date_format
    
    Args:
        date_format (str): the date_format to validate
    
    Returns:
        bool: True if valid, False if not
    """
    if date_format not in ['string_original', 'string_yyyymmdd', 'python_format']:
        raise(ValueError("Invalid date_format. Must be 'string_original', 'string_yyyymmdd', or 'python_format'"))
    else:
        return True
    

def _credentials_validator(creds: credentials.WebserviceCredentials) -> bool:
    """Internal function to validate the credentials
    
    Args:
        credentials (credentials.WebserviceCredentials): the credentials to validate
    
    Returns:
        bool: True if valid, False if not
    """
    if not isinstance(creds, credentials.WebserviceCredentials):
        raise(ValueError("Invalid credentials. Must be a WebserviceCredentials instance"))
    else:
        return True
    
def _session_validator(session: requests.Session) -> bool:
    """Internal function to validate the session
    
    Args:
        session (requests.Session): the session to validate
    
    Returns:
        bool: True if valid, False if not
    """
    if isinstance(session, ffiec_connection.FFIECConnection):
        return True
    elif isinstance(session, requests.Session):
        return True
    else:
        raise(ValueError("Invalid session/connection. Must be a requests.Session instance or FFIECConnection instance"))
    
def _return_client_session(session: requests.Session, creds: credentials.WebserviceCredentials) -> Client:
    
    """Internal function to return a zeep client session
    
    Args:
        session (requests.Session): the requests.Session object to use
        creds (credentials.WebserviceCredentials): the credentials to use

    Returns:
        _type_: _description_
    """
    
    # create a transport
    transport = Transport(session=session)
    
    wsse = UsernameToken(creds.username, creds.password)
                         
    # create the client
    soap_client = Client(constants.WebserviceConstants.base_url, wsse=wsse, transport=transport)
    
    return soap_client




def collect_reporting_periods(session: requests.Session, creds: credentials.WebserviceCredentials, series= "call", output_type = "list", date_output_format="string_original") -> Union[list, pd.Series]:
    """Returns list of reporting periods available for access via the FFIEC webservice

    | Note on `date_output_format`:
    
    * ``string_original`` is the default output format, and is the format that is used by the FFIEC webservice: mm/dd/yyyy
    * ``string_yyyymmdd`` is the date in yyyymmdd format
    * ``python_format`` is the date in python datetime format


    Args:
        session (requests.Session): The requests session object to use for the request.
        creds (credentials.WebserviceCredentials): The credentials to use for the request.
        series (str, optional): `call` or `ubpr`
        output_type (str): `list` or `pandas`
        date_output_format: `string_original`, `string_yyyymmdd`, or `python_format`

    Returns:
        `list` or `Pandas` series: Returns a list of reporting periods from the FFIEC Webservice
        
    """
    
    _ = _output_type_validator(output_type)
    _ = _date_format_validator(date_output_format)
    _ = _credentials_validator(creds)
    _ = _session_validator(session)

    
    ## we have a session and valid credentials, so try to log in
    client  = _client_factory(session, creds)
    
    ## scope ret outside the if statement
    ret = None
    
    if series == "call":
        ret = client.service.RetrieveReportingPeriods(dataSeries="Call")
    elif series == "ubpr":
        ret = client.service.RetrieveUBPRReportingPeriods()
    
    
    # did we return anything? if not, raise an error
    if ret is None:
        raise ValueError("No reporting periods returned.")
    
    ret_date_formatted = ret
        
    if date_output_format == "string_yyyymmdd":
        ret_date_formatted = [datetime.strftime(datetime.strptime(x, "%Y-%m-%d"), "%Y%m%d") for x in ret]
    elif date_output_format == "python_format":
        ret_date_formatted =  [datetime.strptime(x, "%Y-%m-%d") for x in ret]
    # the default is to return the original string
        
    if output_type == "list":
        return ret_date_formatted
    elif output_type == "pandas":
        return pd.DataFrame(ret_date_formatted, columns=['reporting_period'])
    else:
        # for now, default is to return a list
        return ret_date_formatted
    
    pass
    

def _client_factory(session, creds)-> Client:
    """Creates a zeep client session
    
    Determines whether the session argument is an FFIECConnection instance or a requests.Session instance.

    Args:
        session (_type_): _description_
        creds (_type_): _description_

    Returns:
        Client: _description_
    """
        ## we have a session and valid credentials, so try to log in
    if isinstance(session, ffiec_connection.FFIECConnection):
        return  _return_client_session(session.session, creds)
    elif isinstance(session, requests.Session):
        return _return_client_session(session, creds)
    else:
        raise Exception("Invalid session. Must be a FFIECConnection or requests.Session instance")
    

def collect_data(session: Union[ffiec_connection.FFIECConnection, requests.Session], creds: credentials.WebserviceCredentials, reporting_period: Union[str, datetime], rssd_id:str, series: str, output_type = "list", date_output_format ="string_original") -> Union[list, pd.DataFrame]:
    """Return time series data from the FFIEC webservice for a given reporting period and RSSD ID

    Translates the input reporting period to a FFIEC-formatted date
    Transforms the output to a pandas dataframe if output_type is 'pandas', otherwise returns a list
    
    | `Valid arguments for the ``reporting_period`` argument:
    
    * ``mm/dd/yyyy``
    * ``yyyy-mm-dd``
    * ``yyyymmdd``
    *  a python ``datetime`` object
    * For the above types, the date msut be the last day in the quarter (e.g. March 31, June 30, September 30, or December 31)
    * ``#Qyyyy``, where ``#`` is the quarter number and ``yyyy`` is the year.

    Args:
        session (FFIECConnection or requests.Session): The requests session object to use for the request.
        creds (WebserviceCredentials): The credentials to use for the request.
        reporting_period (str or datetime): Reporting period.
        rssd_id (str): The RSSD ID of the entity for which you want to retrieve data.
        series (str): `call` or `ubpr`
        output_type (str): `list` or `pandas`
        date_output_format (str): `string_original`, `string_yyyymmdd`, or `python_format`

    Returns:
        list or pandas: Returns either a list of dicts or a pandas DataFrame
        
    """
    _ = _output_type_validator(output_type)
    _ = _date_format_validator(date_output_format)
    _ = _credentials_validator(creds)
    _ = _session_validator(session)
    

    client = _client_factory(session, creds)
    
    reporting_period_ffiec = _return_ffiec_reporting_date(reporting_period)
    
    #print("Reporting period: {}".format(reporting_period_ffiec))
    
    # try to convert the rssd_id to an int and raise an error if it fails
    rssd_id_int = int(rssd_id)
    
    
    ## scope ret outside the if statement
    ret = None
    
    if series == "call":
        ret = client.service.RetrieveFacsimile(dataSeries="Call", fiIDType="ID_RSSD", fiID=rssd_id_int, reportingPeriodEndDate=reporting_period_ffiec, facsimileFormat="XBRL")
    elif series == "ubpr":
        ret = client.service.RetrieveUBPRXBRLFacsimile(fiIDType="ID_RSSD", fiID=rssd_id_int, reportingPeriodEndDate=reporting_period_ffiec)
    
    processed_ret = xbrl_processor._process_xml(ret, date_output_format)
    
    if output_type == "list":
        return processed_ret
    elif output_type == "pandas":
        return pd.DataFrame(processed_ret)
    
    return processed_ret
    
    
def collect_filers_since_date(session: Union[ffiec_connection.FFIECConnection , requests.Session], creds: credentials.WebserviceCredentials, reporting_period: Union[str, datetime], since_date: Union[str, datetime], output_type = "list") -> Union[list, pd.Series]:
    """Retrieves the ID RSSDs of the reporters who have filed after a given date for a given reporting period. Note that this function only reports on Call Report filings, not UBPR filings.
    
    | `Valid arguments for the ``since_date`` argument:
    
    * ``mm/dd/yyyy``
    * ``yyyy-mm-dd``
    * ``yyyymmdd``
    *  a python ``datetime`` object
    
    | `Valid arguments for the ``reporting_period`` argument:
    
    * all of the above, as long as the date is the last day in the quarter (e.g. March 31, June 30, September 30, or December 31)
    * ``#Qyyyy``, where ``#`` is the quarter number and ``yyyy`` is the year.

    Args:
        session (FFIECConnection or requests.Session): The requests session object to use for the request.
        creds (WebserviceCredentials): The credentials to use for the request.
        since_date (str or datetime): The date to use for the request. May be in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', or a python datetime object.
        output_type (str, optional): "list" or "pandas". Defaults to "list".

    Returns:
        any: Returns either a list of dicts or a pandas Series comprising the ID RSSDs of the reporters who have filed after a given date for a given reporting period.
        
    """
    
    # conduct standard validation on function input arguments
    _ = _output_type_validator(output_type)
    _ = _credentials_validator(creds)
    _ = _session_validator(session)
    
    is_valid_reporting_period = _is_valid_date_or_quarter(reporting_period)
    if not is_valid_reporting_period:
        raise(ValueError("Reporting period must be in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', #QYYYY or a python datetime object, with the month and date set to March 31, June 30, September 30, or December 31."))
    
    
    client = _client_factory(session, creds)
    
    # convert our input dates to the ffiec input date format
    since_date_ffiec = _convert_any_date_to_ffiec_format(since_date)
    reporting_period_datetime_ffiec = _return_ffiec_reporting_date(reporting_period)
    

    
    ret = client.service.RetrieveFilersSinceDate(dataSeries="Call",  lastUpdateDateTime=since_date_ffiec, reportingPeriodEndDate=reporting_period_datetime_ffiec)
    
    
    if output_type == "list":
        return ret
    elif output_type == "pandas":
        return pd.DataFrame(ret, columns=['rssd_id'])
    else:
        # for now, default is to return a list
        return ret
    
    

def collect_filers_submission_date_time(session: Union[ffiec_connection.FFIECConnection, requests.Session], creds: credentials.WebserviceCredentials, since_date: str or datetime, reporting_period: str or datetime, output_type = "list", date_output_format ="string_original") -> Union[list, pd.DataFrame]:
    """Retrieves the ID RSSDs and DateTime of the reporters who have filed after a given date for a given reporting period. Note that this function only reports on Call Report filings, not UBPR filings.

    | Note on `date_output_format`:
    
    * ``string_original`` is the default output format, and is the format that is used by the FFIEC webservice: mm/dd/yyyy
    * ``string_yyyymmdd`` is the date in yyyymmdd format
    * ``python_format`` is the date in python datetime format

    Args:
        session (ffiec_connection.FFIECConnection or requests.Session): The requests session object to use for the request.
        creds (WebserviceCredentials or requests.Session): The credentials to use for the request.
        since_date (str or datetime): The date to use for the request. May be in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', or a python datetime object.
        reporting_period (str or datetime): The reporting period to use for the request (e.g. "2020-03-21"). Note that the date must be in the format of "YYYY-MM-DD", "YYYYMMDD", "MM/DD/YYYY", #QYYYY or a python datetime object, with the month and date set to March 31, June 30, September 30, or December 31.
        output_type (str, optional): "list" or "pandas". Defaults to "list".
        date_output_format (str, optional): string_original or python_datetime. Defaults to "string_original".

    Returns:
        any: List of dicts or pandas DataFrame containing the rssd id of the filer, and the submission date and time, in Washington DC timezone.
    """
    
        
    # conduct standard validation on function input arguments
    _ = _output_type_validator(output_type)
    _ = _date_format_validator(date_output_format)
    _ = _credentials_validator(creds)
    _ = _session_validator(session)
    
    is_valid_reporting_period = _is_valid_date_or_quarter(reporting_period)
    if not is_valid_reporting_period:
        raise(ValueError("Reporting period must be in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', #QYYYY or a python datetime object, with the month and date set to March 31, June 30, September 30, or December 31."))
    
    
    ## we have a session and valid credentials, so try to log in
    client = _client_factory(session, creds)
    
    # convert our input dates to the ffiec input date format
    since_date_ffiec = _convert_any_date_to_ffiec_format(since_date)
    reporting_period_datetime_ffiec = _return_ffiec_reporting_date(reporting_period)
    
    # send the request
    
    # first, create the client
    client = _client_factory(session, creds)
    
    ret = client.service.RetrieveFilersSubmissionDateTime(dataSeries="Call", lastUpdateDateTime=since_date_ffiec, reportingPeriodEndDate=reporting_period_datetime_ffiec)
    
    # normalize the output
    normalized_ret = [{"rssd":x["ID_RSSD"], "datetime":x["DateTime"]} for x in ret]

    # all submission times are in eastern time, so if we are converting to a python datetime,
    # the datetime object needs to be timezone aware, so that the user may convert the time to their local timezone    
    origin_tz = ZoneInfo("US/Eastern")

    
    if date_output_format == "python_format":
        normalized_ret = [{"rssd":x["rssd"], "datetime":datetime.strptime(x["datetime"], "%m/%d/%Y %H:%M:%S %p").replace(tzinfo=origin_tz)} for x in normalized_ret]
    

    # convert the datetime to a string, if user requests
    
    if output_type == "list":
        return normalized_ret
    elif output_type == "pandas":
        return pd.DataFrame(normalized_ret)
    else:
        # for now, default is to return a list
        return ret
    
    
    pass

def collect_filers_on_reporting_period(session: Union[ffiec_connection.FFIECConnection, requests.Session], creds: credentials.WebserviceCredentials, reporting_period: Union[str, datetime], output_type = "list") -> Union[list, pd.DataFrame]:
    """Retrieves the Financial Institutions in a Panel of Reporters for a given reporting period. Note that this function only reports on Call Report filings, not UBPR filings.

    | `Valid arguments for the ``reporting_period`` argument:
    
    * ``mm/dd/yyyy``
    * ``yyyy-mm-dd``
    * ``yyyymmdd``
    *  a python ``datetime`` object
    * For the above types, the date must be the last day in the quarter (e.g. March 31, June 30, September 30, or December 31)
    * ``#Qyyyy``, where ``#`` is the quarter number and ``yyyy`` is the year.


    Args:
        session (ffiec_connection.FFIECConnection or requests.Session): The requests session object to use for the request.
        creds (credentials.WebserviceCredentials): The credentials to use for the request.
        reporting_period (str or datetime): The reporting period to use for the request.
    Returns:
        list or pd.DataFrame: List of dicts or pandas DataFrame containing the rssd id of the filer, and the following fields: "id_rssd", "fdic_cert_number", "occ_chart_number", "ots_dock_number", "primary_aba_rout_number", "name", "state", "city", "address", "filing_type", "has_filed_for_reporting_period"
    """
    
        # conduct standard validation on function input arguments
    _ = _output_type_validator(output_type)
    _ = _credentials_validator(creds)
    _ = _session_validator(session)
    
    is_valid_reporting_period = _is_valid_date_or_quarter(reporting_period)
    if not is_valid_reporting_period:
        raise(ValueError("Reporting period must be in the format of 'YYYY-MM-DD', 'YYYYMMDD', 'MM/DD/YYYY', #QYYYY or a python datetime object, with the month and date set to March 31, June 30, September 30, or December 31."))
    
    
    client = _client_factory(session, creds)
    reporting_period_datetime_ffiec = _return_ffiec_reporting_date(reporting_period)
    
    ret = client.service.RetrievePanelOfReporters(dataSeries="Call", reportingPeriodEndDate=reporting_period_datetime_ffiec)
    
    normalized_ret = [datahelpers._normalize_output_from_reporter_panel(x) for x in ret]
    
    if output_type == "list":
        return normalized_ret
    elif output_type == "pandas":
        return pd.DataFrame(normalized_ret)
    else:
        # for now, default is to return a list
        return ret
    



"""Methods to utilize for inputting credentials for the FFIEC data connection.

This module provides secure methods for inputting credentials for the FFIEC webservice data connection.

Credentials may be input via environment variables, or passing them as arguments into the class structure. Wherever possible, the credentials should not be stored in source code.

"""

import requests
import sys
import os

from enum import Enum

from zeep import Client, Settings
from zeep.wsse.username import UsernameToken
from zeep.transports import Transport

from ffiec_data_connect import constants, ffiec_connection

class CredentialType(Enum):
    """Enumerated values that represent the methods through which credentials are provided to the FFIEC webservice via the package.

    Args:
        Enum (integer): Integer that represents the credential input method
    """
    NO_CREDENTIALS = 0
    SET_FROM_INIT = 1
    SET_FROM_ENV = 2

class WebserviceCredentials(object):
    """The WebserviceCredentials class. This class is used to store the credentials for the FFIEC webservice.
    
    Args:
        username (str, optional): FFIEC Webservice username. Optional: If not provided, the credentials will be set from the environment variable `FFIEC_USERNAME`
        password (str, optional): FFIEC Webservice password. Optional: If not provided, the credentials will be set from the environment variable `FFIEC_PASSWORD`
    
    Returns:
        WebserviceCredentials: An instance of the WebserviceCredentials class.
    
    """
    
    def __init__(self, username = None, password = None):

        # collect the credentials from the environment variables
        # if the environment variables are not set, we will set the credentials from the arguments
        username_env = os.getenv("FFIEC_USERNAME")
        password_env = os.getenv("FFIEC_PASSWORD")

        # if we are passing in credentials, use them
        if password and username:
            self.username = username
            self.password = password
            self.credential_source: CredentialType = CredentialType.SET_FROM_INIT
            return
        
        # if not, check if we have the two environment variables
        
        # do we have both environment variables?
        elif username_env and password_env:
            self.username = username_env
            self.password = password_env
            self.credential_source: CredentialType = CredentialType.SET_FROM_ENV
        
        else:
            # do we have a username and password?
            self.credential_source = CredentialType.NO_CREDENTIALS

            raise ValueError("Username and password must be set to create a connection")
        
        return
    
        
    def __str__(self) -> str:
        """String representation of the credentials."""
        if self.credential_source  == CredentialType.NO_CREDENTIALS or self.credential_source == None:
            return "No credentials set"
        elif self.credential_source == CredentialType.SET_FROM_INIT:
            return f"Credentials set from class initialization with username: {self.username}"         
        elif self.credential_source == CredentialType.SET_FROM_ENV:
            return f"Credentials set from environment variables with username: {self.username}"            
        else:
            return "Unknown credential source"
    
    def __repr__(self) -> str:
        return self.__str__()

        
    def test_credentials(self, session: requests.Session) -> bool:
        """Test the credentials with the FFIEC Webservice to determine if they are valid and accepted.
        
            | Note: The session argument can be generated directly from requests, or
            | using the helper class `FFIECConnection`
            
            
        Args:
            session (requests.Session): the connection to test the credentials 
        
        Returns:
            bool: True if the credentials are valid, False otherwise
            
        Raises:
            ValueError: if the credentials are not set
            Exception: Other unspecified errors
            
        """
    
        # check that we have a user name
        if self.username is None:
            raise ValueError("Username must be set")
        
        # check that we have a password
        if self.password is None:
            raise ValueError("Password must be set")
        
        # we have a user name and password, so try to log in
        # create the token
        wsse = UsernameToken(self.username, self.password)

        try:

            # create a transport
            transport = None
            
            if isinstance(session, requests.Session):
                transport = Transport(session=session)                
            elif isinstance(session, ffiec_connection.FFIECConnection):
                transport = Transport(session=session.session)
            
            # create the client
            soap_client = Client(constants.WebserviceConstants.base_url, wsse=wsse, transport=transport)
            
            print("Standby...testing your access.")    
            
            has_access_response = soap_client.service.TestUserAccess()
            
            if has_access_response:
                print("Your credentials are valid.")
                return
            else:
                print("Your credentials are invalid. Please refer to the documentation for more information.")
                print(has_access_response)
                return
                
            print(has_access_response)
                
                        
        except Exception as e:
            print(
                "Credentials are invalid. Please check documentation for more information."
            )
            print("Credentials error: {}".format(e))
            
            raise(Exception("Credentials are invalid. Please check documentation for more information."))
        
        
    
    @property
    def username(self) -> str:
        """Returns the username from the WebserviceCredentials instance.
        
        Returns:
            str: the username stored in the WebserviceCredentials instance
        """
        return self._username

    @username.setter
    def username(self, username: str) -> None:
        """Sets the username in the WebserviceCredentials instance.
        
        Args:
            username (str): the username to set in the WebserviceCredentials instance
        """
 
        self._username = username
        pass
    
    @property
    def password(self) -> str:
        """Returns the password from the WebserviceCredentials instance.
        
        Returns:
            str: the password stored in the WebserviceCredentials instance
        """
        return self._password
    
    
    @password.setter
    def password(self, password: str) -> None:
        """Sets the password in the WebserviceCredentials instance.
        
        Args:
            password (str): the password to set in the WebserviceCredentials instance
        """
        self._password = password
        pass



"""Internal functions used to process XBRL data received from the FFIEC Webservice
"""
from itertools import chain
import xmltodict
from datetime import datetime
import re

re_date = re.compile('[0-9]{4}\-[0-9]{2}\-[0-9]{2}')

def _process_xml(data: bytes, output_date_format: str):
    #data = zipfile_stream.open(first_file).read()
    dict_data = xmltodict.parse(data.decode('utf-8'))['xbrl']

    keys_to_parse = list(filter(lambda x: 'cc:' in x, dict_data.keys())) + list(filter(lambda x: 'uc:' in x, dict_data.keys()))
    parsed_data = list(chain.from_iterable(filter(None,list(map(lambda x: _process_xbrl_item(x, dict_data[x], output_date_format),keys_to_parse,)))))
    ret_data = []
    for row in parsed_data:
        new_dict = {}
        new_dict.update({'mdrm':row['mdrm']})
        new_dict.update({'rssd':row['rssd']})
        new_dict.update({'quarter':row['quarter']})
        if row['data_type'] == 'int':
            new_dict.update({'int_data':int(row['value'])})
            new_dict.update({'float_data':None})
            new_dict.update({'bool_data':None})
            new_dict.update({'str_data':None})
            new_dict.update({'data_type':row['data_type']})

        elif row['data_type'] == 'float':
            new_dict.update({'int_data':None})
            new_dict.update({'float_data':row['value']})
            new_dict.update({'bool_data':None})
            new_dict.update({'str_data':None})
            new_dict.update({'data_type':row['data_type']})

        elif row['data_type'] == 'str':
            new_dict.update({'int_data':None})
            new_dict.update({'float_data':None})
            new_dict.update({'bool_data':None})
            new_dict.update({'str_data':row['value']})
            new_dict.update({'data_type':row['data_type']})

        elif row['data_type'] == 'float':
            new_dict.update({'int_data':None})
            new_dict.update({'float_data':row['value']})
            new_dict.update({'bool_data':None})
            new_dict.update({'data_type':row['data_type']})
            new_dict.update({'str_data':None})

        elif row['data_type'] == 'bool':
            new_dict.update({'int_data':None})
            new_dict.update({'float_data':None})
            new_dict.update({'bool_data':row['value']})
            new_dict.update({'data_type':row['data_type']})
            new_dict.update({'str_data':None})

        ret_data.append(new_dict)
    
    return ret_data


def _create_ffiec_date_from_datetime(indate: datetime) -> str:
    """Converts a datetime object to a FFIEC-formatted date

    Args:
        indate (datetime): the date to convert

    Returns:
        str: the date in FFIEC format
    """
    month_str = str(indate.month)
    day_str = str(indate.day)
    year_str = str(indate.year)
    
    mmddyyyy = month_str + "/" + day_str + "/" + year_str
    
    return mmddyyyy

def _process_xbrl_item(name, items, date_format):
    # incoming is a data dictionary
    results = []
    if type(items) != list:
        items = [items]
    for j,item in enumerate(items):
        context = item.get('@contextRef')
        unit_type = item.get('@unitRef')
        value = item.get('#text')
        mdrm = name.replace("cc:","").replace("uc:","")
        rssd = context.split('_')[1]
        #date = int(context.split('_')[2].replace("-",''))

        quarter = re_date.findall(context)[0]

        # transform the date to the requested date format
        if date_format == 'string_original':
            quarter = _create_ffiec_date_from_datetime(datetime.strptime(quarter, '%Y-%m-%d'))
        elif date_format == 'string_yyyymmdd':
            quarter = datetime.strptime(quarter, '%Y-%m-%d').strftime('%Y%m%d')
        elif date_format == 'python_format':
            quarter = datetime.strptime(quarter, '%Y-%m-%d')
        
        data_type = None


        if unit_type == 'USD':
            value = int(value)/1000
            data_type = 'int'
        elif unit_type == 'PURE':
            value = float(value)
            data_type = 'float'
        elif unit_type == 'NON-MONETARY':
            value = float(value)
            data_type = 'float'
        elif value == 'true':
            value = True
            data_type = 'bool'
        elif value == 'false':
            value = False
            data_type = 'bool'
        else:
            data_type = 'str'                

        results.append({'mdrm':mdrm, 'rssd':rssd, 'value':value, 'data_type':data_type, 'quarter':quarter})

    return results



"""Wrapper to establish requests.Session and set proxy server values if needed

`An instance of this class may be substituted for an instance requests.Session class.`
"""

import requests
from enum import Enum


class ProxyProtocol(Enum):
    """Enumerated values that represent the proxy protocol options

    Args:
        Enum (ProxyProtocol): Enumerated value for HTTP or HTTPS
    """
    
    HTTP = 0
    HTTPS = 1

class FFIECConnection(object):
    """Creates a FFIECConnection object, which may include proxy server connection parameters
    """
    
    def __init__(self) -> None:
        """Initializes the Https Connection to be utilized
        to connect to the FFIEC website
        
        Args:
            None
        
        
        """
        
        self.use_proxy = False
        self.proxy_host = None
        self.proxy_port = None
        self.proxy_password = None
        self.proxy_user_name = None
        self.proxy_protocol = None
        
        self._generate_session()
        
        return
    
    @property
    def session(self) -> requests.Session:
        """Returns the requests.Session object
        * Note that this property may be utilized for methods in the `methods` module.
        
        `This session property is automatically generated when the FFIECConnection object is created.`
        
        Returns:
            requests.Session: the requests.Session object
        
        """
        return self._session
    
    @session.setter
    def session(self, session: requests.Session) -> None:
        """Set the requests.Session object
        
        Args:
            session (requests.Session): the requests.Session object
        """
        self._session = session
        return
        
    @property
    def proxy_host(self) -> str:
        """Returns the hostname of the proxy server
        
        Returns:
            str: The hostname of the proxy server
        
        """
        return self._proxy_host
        
    @proxy_host.setter
    def proxy_host(self, host: str) -> None:
        """Set the optional proxy hostname

        Args:
            host (str): the host name of the proxy server
        """
        
        self._generate_session()
        
        self._proxy_host = host
        pass
    
    @property
    def proxy_protocol(self) -> int:
        """Returns the protocol of the proxy server
        
        Returns:
            int: the protocol of the proxy server
        
        """
        return self._proxy_protocol
    
    @proxy_protocol.setter
    def proxy_protocol(self, protocol: ProxyProtocol) -> None:
        """Set the optional proxy protocol

        Args:
            protocol (str): the protocol of the proxy server
        """
        self._generate_session()
        
        self._proxy_protocol = protocol
        pass
    
    @property
    def proxy_port(self) -> int:
        """Get the optional proxy port

        Returns:
            int: the proxy port
        """
        return self._proxy_port
    
    @proxy_port.setter
    def proxy_port(self, port: int) -> None:
        """Set the optional proxy port

        Args:
            port (int): the port of the proxy server
        """
        
        self._generate_session()
        
        self._proxy_port = port
        pass
    
    @property
    def proxy_user_name(self) -> str:
        """Get the optional proxy username
        
        Returns:
            str: the proxy username
        
        """
        
        self._generate_session()
        
        return self._proxy_user_name
    
    @proxy_user_name.setter
    def proxy_user_name(self, username: str) -> None:
        """Set the optional proxy username

        Args:
            username (str): the username of the proxy server
        """
        
        self._generate_session()
        
        self._proxy_user_name = username
        pass
    
    @property
    def proxy_password(self) -> str:
        """Get the optional proxy password
        
        Returns:
            str: the password of the proxy server
        
        """
        return self._proxy_password
    
    @proxy_password.setter
    def proxy_password(self, password: str) -> None:
        """Set the optional proxy password

        Args:
            password (str): the password of the proxy server
        """
        
        self._generate_session()
        
        self._proxy_password = password
        pass
    
    @property
    def use_proxy(self) -> bool:
        """Get the optional proxy flag
        
        Returns:
            bool: the proxy flag (True if proxy is used, False if not)
        
        """
        return self._use_proxy
    
    @use_proxy.setter
    def use_proxy(self, use_proxy_opt: bool) -> None:
        """Set the optional proxy flag
        If set to True, the proxy server will be used

        Args:
            useProxy (bool): the flag to use the proxy server
        """
        
        # # if we are setting the proxy to true, check that we have, at minimum, a host name, port, and protocol
        # if use_proxy_opt:
        #     if self.proxy_host is None or self.proxy_port is None or self.proxy_protocol is None:
        #         raise ValueError("Proxy host, port, and protocol must be set before using the proxy")
        
        #     else:
        #         self._use_proxy = use_proxy_opt
                
        #         self._generate_session()
        # else:
        #     self._generate_session()

        self._use_proxy = use_proxy_opt
        
        self._generate_session()
            
        return
    
   
    def _generate_session(self) -> requests.Session:
        """Internal class method to generate a requests session object

        """
        
        # create a requests session
        session = requests.Session()
        
        # are we using a proxy?
        if self.use_proxy:
            
            # check that we have a hostname, port, and protocol. If not, raise an error
            if self.proxy_host is None or self.proxy_port is None or self.proxy_protocol is None:
                raise()
            
            
            # if we are using a proxy, set the proxy host and port
            session.proxies = {self.proxy_protocol.name: 'http://' + self.proxy_host + ':' + str(self.proxy_port)}
            # if we have a username and password, set the proxy username and password
            if self.proxy_user_name is not None and self.proxy_password is not None:
                session.proxies['http']['proxy_auth'] = (self._proxy_user_name, self.proxy_password)
        
         # set the session to self.connection
        self.session = session

        return
    
   
    def test_connection(self, url: str = "https://google.com") -> bool:
        """Tests a connection, using the proxy server if one is not set
        
        Note: This method tests if a web page on the public internet (not the FFIEC webservice) 
        is accessible using this library, through the proxy server.
        
        We do not yet test access to the Webservice, because a user name and password
        is needed to access the Webservice, which is outside the scope of this module.
        
        An alternative web site may be selected in lieu of google.com,
        using the url parameter.
        
        """
        
        # can we return a HTTP 200 from google.com, or the url specified??
        
        if url is None:
            url = "https://google.com"
        
        response = self.session.get(url)
        
        if response.status_code == 200:
            return True
        else:
            print("Unable to access test site via proxy. Error: " + str(response.status_code))
            return False


    def __str__(self) -> str:
        
        """String representation of the HttpsConnection object
        
        Returns:
            str: the string representation of the HttpsConnection object
        
        """
        return f"""
        HttpsConnection object properties:
        
        Https connection session is {'active' if self.session is not None else 'inactive'}
        
        
        proxy hostname = {self.proxy_host}
        proxy port = {self.proxy_port}
        proxy protocol = {self.proxy_protocol}
        is the proxy active? = {self.use_proxy}
        proxy username = {self.proxy_user_name}
        is proxy password set? = {self.proxy_password is not None}
        
        """
        
    

    def __repr__(self) -> str:
        return self.__str__()


"""Internal functions to assist with processing results from the Webservice
"""


from zeep import helpers

def _normalize_output_from_reporter_panel(row: dict) -> dict:

    """Normalize the output from the reporter panel into a dict
    Converts integers to strings,
    for zipcode, convert to string and zfill for 5 digits
    change hasFiledForReportingPeriod to bool,
    and change all field names to camel case

    Returns:
        _type_: _description_
    """



    new_row = {}
    row_keys = list(helpers.serialize_object(row).keys())
    
    # process ID_RSSD
    if 'ID_RSSD' in row_keys:
        new_row['id_rssd'] = str(row['ID_RSSD'])
    else:
        new_row['id_rssd'] = None

    # process FDIC CERT NUMBER
    if 'FDICCertNumber' in row_keys:
        if row['FDICCertNumber'] != 0:
            new_row['fdic_cert_number'] = str(row['FDICCertNumber'])
        else:
            new_row['fdic_cert_number'] = None
    else:
        new_row["fdic_cert_number" ] = None
        
    # OCC chart number
    if 'OCCChartNumber' in row_keys:
        if row['OCCChartNumber'] != 0:
            new_row['occ_chart_number'] = str(row['OCCChartNumber'])
        else:
            new_row['occ_chart_number'] = None
    else: 
        new_row['occ_chart_number'] = None
        
    # process OTSDockNumber
    if 'OTSDockNumber' in row_keys:
        if row['OTSDockNumber'] != 0:
            new_row['ots_dock_number'] = str(row['OTSDockNumber'])
        else:
            new_row['ots_dock_number'] = None
    else:
        new_row['ots_dock_number'] = None
        
    # process PrimaryABARoutNumber
    if 'PrimaryABARoutNumber' in row_keys:
        if row['PrimaryABARoutNumber'] != 0:
            new_row['primary_aba_rout_number'] = str(row['PrimaryABARoutNumber'])
        else:
            new_row['primary_aba_rout_number'] = None
    
    # Process Name
    if 'Name' in row_keys:
        temp_str = row['Name'].strip()
        if temp_str == '0' or temp_str == '':
            new_row['name'] = temp_str
        else:
            new_row['name'] = temp_str
       
    else:
        new_row['name'] = None
    
    #Process State
    if 'State' in row_keys:
        temp_str = row['State'].strip()
        if temp_str == '0' or temp_str == '':
            new_row['state'] = temp_str
        else:
            new_row['state'] = temp_str
    else:
        new_row['city'] = None
    
    # Process City
    if 'City' in row_keys:
        temp_str = row['City'].strip()
        if temp_str == '0' or temp_str == '':
            new_row['city'] = temp_str
        else:
            new_row['city'] = temp_str        
    else:
        new_row['city'] = None

    
    # Process Address
    if 'Address' in row_keys:
        temp_str = row['Address'].strip()
        if temp_str == '0' or temp_str == '':
            new_row['address'] = temp_str
        else:
            new_row['address'] = temp_str
            

    #Process Zip
    if 'Zip' in row_keys:
        temp_str = str(row['Zip']).zfill(5)
        if temp_str == '0' or temp_str == '':
            new_row['zip'] = temp_str
        else:
            new_row['zip'] = temp_str
            
    
    #Process FilingType
    if 'FilingType' in row_keys:
        temp_str = row['FilingType'].strip()
        if temp_str == '0' or temp_str == '':
            new_row['filing_type'] = temp_str
        else:
            new_row['filing_type'] = temp_str
            
    # process HasFiledForReportingPeriod
    if 'HasFiledForReportingPeriod' in row_keys:
        if type(row['HasFiledForReportingPeriod']) == bool:
            new_row['has_filed_for_reporting_period'] = row['HasFiledForReportingPeriod']
        else:
            new_row['has_filed_for_reporting_period'] = None
            

    return new_row
    
    


"""Constant values utilized for data collection and other purposes.

This module contains constant values that are unlikely to change,
but need to be referenced by other modules.

"""

class WebserviceConstants(object):
    
    """The URL endpoint for the FFIEC SOAP-based webservice.
    """
    base_url = "https://cdr.ffiec.gov/Public/PWS/WebServices/RetrievalService.asmx?WSDL"




from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='ffiec-data-connect',
    python_requires='>3.9.0',
    version='0.2.7',
    license='MIT',
    description="Wrapper for the FFIEC's Webservice API",
    readme='README.md',
    long_description=long_description,
    project_urls={
    'Documentation': 'https://ffiec-data-connect.readthedocs.io/en/latest/',
    'Repo': 'https://github.com/call-report/ffiec-data-connect',
    'Additional Info': 'http://call.report',
    'Author': 'https://mikeh.dev',
},
    long_description_content_type='text/markdown',
    author="Michael Handelman",
    author_email='m@mikeh.dev',
    packages=find_packages('src'),
    include_package_data=True,
    package_dir={'': 'src'},
    url='https://github.com/call-report/ffiec-data-connect',
    keywords='ffiec call report bank regulatory',
    install_requires=[
          'zeep',
          'xmltodict',
          'requests',
          'pandas'
      ],

)